#include "scriptparser.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "pixel_formats.h"


int32 CScriptParser::m_LineNumber;


/////////////////////////////////////////////////////////////////////////////////
//////////////////////// Public functions (interface) ///////////////////////////
/////////////////////////////////////////////////////////////////////////////////


bool CScriptParser::ParseScript(const char pScriptPath[])
{
	// fetch file
	bool bSuccess = false;
	FILE * fptr = NULL;
	if((fptr=fopen(pScriptPath, "rb")) != NULL)
	{
		// determine size of script
		fseek(fptr, 0, SEEK_END);
		const int32 end = ftell(fptr);
		fseek(fptr, 0, SEEK_SET);
		szeScript = end - ftell(fptr);
		assert(szeScript > 0);

		// Read scriptfile
		if((pScript=new char[szeScript]) != NULL)
		{
			fread(pScript, szeScript, 1, fptr);
			//m_bValid=true;

			// setup token parsing
			token.cptr = pScript;
			token.str_len = 0;
			token.token_type = UNKNOWN;

			// run script
			bSuccess = PriScriptParse();
			delete [] pScript; pScript = NULL;
		}
		else
		{
			fclose(fptr);
			printf("Memory allocation error in scriptparse: %s");
		}

		// close file
		fclose(fptr); fptr = NULL;
	}
	else
		printf("Couldn't open script file %s\n", pScriptPath);

	// warnings and comments
	if(m_cInputContainer.GetOutputFolder() == NULL)
	{
		printf("Warning! Found now specified outfolder in script.\n");
		bSuccess = false;
	}
	if(m_cInputContainer.GetNum3dsInputFiles() == 0)
	{
		printf("Warning! No input files specified in script.\n");
		bSuccess = false;
	}
	
	return bSuccess;
}

const CInputContainer CScriptParser::GetInputContainer()
{
	return m_cInputContainer;
}


CScriptParser::CScriptParser()
{



}

CScriptParser::~CScriptParser()
{



}



/////////////////////////////////////////////////////////////////////////////////
////////////////////////////// Parse Functions //////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

bool CScriptParser::PriScriptParse()
{
	bool bTokensLeft = true;
	m_LineNumber=1;

	while(bTokensLeft)
	{
		// get us a valid Token
		while(bTokensLeft && token.token_type == UNKNOWN)
			bTokensLeft = match(token.token_type);

		switch(token.token_type)
		{
			case INPUT_FILE:
				bTokensLeft = ParsePath(INPUT_FILE);
				break;
			case INPUT_FILES:
				bTokensLeft = ParsePaths(INPUT_FILES);
				break;
			case TEXTURE_FOLDER:
				bTokensLeft = ParsePath(TEXTURE_FOLDER);
				break;
			case TEXTURE_FOLDERS:
				bTokensLeft = ParsePaths(TEXTURE_FOLDERS);
				break;
			case STRIP_MAXVERTS:
				bTokensLeft = ParseStripMaxVerts();
				break;
			case OUTPUT_FOLDER:
				bTokensLeft = ParseOutputFolder();
				break;
			case TEXLIST:
				bTokensLeft = ParseListOfSpecificTexProps();
				break;
			case TEXDEFAULT:
				bTokensLeft = ParseDefaultTexProps();
				break;
			default:
				PrintError("Unknown token");
				assert(false);
		}
	}

	return true;
}


bool CScriptParser::ParsePaths(const ASC_TOKEN type)
{
	match(type);
	match(EQUAL);
	match(LEFT_CURLY);
	
	// fetch all strings
	// parse token string
	bool bTokensLeft = true;
	while(bTokensLeft && token.token_type!=RIGHT_CURLY)
	{
		if(token.token_type != STRING)
			PrintError("Expecting string");
		assert(token.token_type == STRING);

		// add current string to list of paths
		bool res=false;
		if(type == INPUT_FILES)
			res = m_cInputContainer.AddInputFile(token.cptr, token.str_len);
		else if(type == TEXTURE_FOLDERS)
			res = m_cInputContainer.AddTextureFolder(token.cptr, token.str_len);
		assert(res);

		// next token
		bTokensLeft = match(STRING);
		if(token.token_type == COMMA)
		{
			bTokensLeft = match(COMMA);
			assert(token.token_type == STRING);
		}
	}

	// close
	return match(RIGHT_CURLY);
}

bool CScriptParser::ParsePath(const ASC_TOKEN type)
{
	match(type);
	match(EQUAL);
	
	if(token.token_type != STRING)
		PrintError("Expecting string");
	assert(token.token_type == STRING);

	// fetch string
	bool res = false;
	if(type == INPUT_FILE)
		res = m_cInputContainer.AddInputFile(token.cptr, token.str_len);
	else if(type == TEXTURE_FOLDER)
		res = m_cInputContainer.AddTextureFolder(token.cptr, token.str_len);
	assert(res);
	
	return match(STRING);
}

bool CScriptParser::ParseOutputFolder()
{
	match(OUTPUT_FOLDER);
	match(EQUAL);

	if(token.token_type != STRING)
		PrintError("Expecting string");
	assert(token.token_type == STRING);

	// set folder
	bool res = m_cInputContainer.SetOutputFolder(token.cptr, token.str_len);
	assert(res);

	return match(STRING);
}

bool CScriptParser::ParseStripMaxVerts()
{
	match(STRIP_MAXVERTS);
	match(EQUAL);

	if(token.token_type != INTEGER)
		PrintError("Expecting integer");
	assert(token.token_type == INTEGER);

	// set value
	m_cInputContainer.SetStripMaxVerts( token_to_integer() );

	return match(INTEGER);
}

bool CScriptParser::ParseListOfSpecificTexProps()
{
	match(TEXLIST);
	match(EQUAL);
	match(LEFT_CURLY);
	
	// fetch all props
	// parse token prop
	bool bTokensLeft = true;
	while(bTokensLeft && token.token_type!=RIGHT_CURLY)
	{
		match(TEX);

		// get prop
		CTextureProps texProps;
		bool bTokensLeft = ParseTextureProps(false, &texProps);
		bool res = m_cInputContainer.AddSpecificTextureProps( texProps );
		assert( res );

		// next token
		if(token.token_type == COMMA)
		{
			bTokensLeft = match(COMMA);
			if(token.token_type != TEX)
				PrintError("Expecting a tex token");
			assert(token.token_type == TEX);
		}
	}

	// close
	return match(RIGHT_CURLY);
}

bool CScriptParser::ParseDefaultTexProps()
{
	match(TEXDEFAULT);
	match(EQUAL);

	CTextureProps texProps;
	bool bTokensLeft = ParseTextureProps(true, &texProps);
	m_cInputContainer.SetDefaultTextureProps( texProps ); 
	return bTokensLeft;
}



bool CScriptParser::ParseTextureProps(const bool bIsDefaultProps, CTextureProps * pTexProps)
{
	// begin
	bool bTokensLeft = match(LEFT_BRACKET);

	// parse inbetween props
	CTextureProps &texProps = *pTexProps;

	// look for name
	if(!(token.token_type==STRING || bIsDefaultProps))
		PrintError("Expecting string");
	assert(token.token_type==STRING || bIsDefaultProps);
	if(token.token_type == STRING)
	{
		texProps.SetTexPath(token.cptr, token.str_len);
		match(STRING);
		bTokensLeft = match(COMMA);
	}

	// Parse through prop args
	while(bTokensLeft && token.token_type!=RIGHT_BRACKET && token.token_type!=STRING)
	{
		switch(token.token_type)
		{
			// output depths for textures
			case PACK_INDEX16:
				texProps.SetMode(INDEXCOLOR16);
				break;
			case PACK_INDEX256:
				texProps.SetMode(INDEXCOLOR256);
				break;
			case PACK_RGBA32:
				texProps.SetMode(COLORRGBA32);
				break;
			
			// request generation of mipmap levels
			case PACK_NOMIPMAP:
				texProps.SetMipMap(false);
				break;
			case PACK_MIPMAP:
				texProps.SetMipMap(true);
				break;

			// set dithering properties
			case PACK_NODITHER:
				texProps.SetDither(false);
				break;
			case PACK_DITHER:
				texProps.SetDither(true);
				break;
			default:
				PrintError("Unknown pack instruction");
				assert(false);
		}
		
		// next token, finish current prop
		match(token.token_type);
		if(token.token_type == COMMA)
		{
			bTokensLeft = match(COMMA);
			if(token.token_type!=STRING)
				PrintError("Expecting string");
			assert(token.token_type == STRING);
		}
		else if(token.token_type!=RIGHT_BRACKET)
			match(PIPE);
	}

	// Pickup user defined properties if any were found
	if(token.token_type == STRING)
	{
		bool res = texProps.SetUserdefinedProps(token.cptr, token.str_len);
		assert(res);
		match(STRING);
	}


	// end
	return match(RIGHT_BRACKET);
}



/////////////////////////////////////////////////////////////////////////////////
///////////////////////////// lexical analysis //////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

char CScriptParser::token_strings[][MAX_TOKEN_STRING_LENGTH] =
{
	{"<"},
	{">"},
	{"{"},
	{"}"},
	{"["},
	{"]"},
	{","},
	{":"},
	{""},	// empty for string

	{""},	// empty for integer
	{""},	// empty for real_value

	{"="},
	{"|"},
	{"input_files"},
	{"input_file"},
	{"strip_maxverts"},
	{"texture_folders"},
	{"texture_folder"},
	{"output_folder"},
	{"texdefault"},
	{"texlist"},
	{"tex"},
	
	{"PACK_INDEX16"},
	{"PACK_INDEX256"},
	{"PACK_RGBA32"},
	{"PACK_NOMIPMAP"},
	{"PACK_MIPMAP"},
	{"PACK_NODITHER"},
	{"PACK_DITHER"}
};

// globals
int32 CScriptParser::szeScript = 0;
char * CScriptParser::pScript = NULL;
CScriptParser::Token CScriptParser::token;


bool CScriptParser::match(ASC_TOKEN type)
{
	if(token.token_type != type)
		PrintError("Expected a different token");
	assert(token.token_type == type);

	// get next token
	return fetch_token();
}

bool CScriptParser::fetch_token()
{
	bool res = false;

	// offset by previous string length, plus one to
	// remove '"' if last token was a string
	token.cptr += token.str_len;
	if(token.token_type == STRING)
		++token.cptr;

	// Remove whitespace
	RemoveWhiteSpace();

	// if we're left with any chars
	if(token.cptr < (pScript+szeScript))
	{
		if(token.cptr[0]=='"')	// check to see if this might be of type string
		{
			++token.cptr;	// skip '"'
			assert( token.cptr < (pScript+szeScript)/* && token.cptr[0]!='"'*/);
			int n=0;
			while(token.cptr[n]!='"')
			{
				++n;
				assert( (token.cptr+n) < (pScript+szeScript));
			}
			
			token.str_len = n;
			token.token_type = STRING;
			res = true;
		}	// fetch integer or real_value
		else if(isdigit(token.cptr[0]) || (token.cptr[0]=='-' && (token.cptr+1<(pScript+szeScript)) && isdigit(token.cptr[1])))
		{
			int n=0;
			if( ((token.cptr+n) < (pScript+szeScript)) && token.cptr[n]=='-')
				++n;
			while( ((token.cptr+n) < (pScript+szeScript)) && isdigit(token.cptr[n]) )
				++n;

			// integer
			if( ((token.cptr+n) >= (pScript+szeScript)) || token.cptr[n]!='.')
			{
				token.str_len = n;
				token.token_type = INTEGER;
				res = true;
			}
			else	// real_value
			{
				++n;	// skip the dot
				assert( ((token.cptr+n) < (pScript+szeScript)) && isdigit(token.cptr[n]));
				while( ((token.cptr+n) < (pScript+szeScript)) && isdigit(token.cptr[n]) )
					++n;

				token.str_len = n;
				token.token_type = REAL_VALUE;
				res = true;
			}
		}
		else	// test all non-empty token strings for a match
		{
			int32 matchedIndex=-1, matchedLen = 0;
			for(int i=0; i<NUM_OF_ASC_TOKENS; i++)
			{
				// strings such as those of 
				// real_value and of the "symbol string"
				// are left empty
				int32 strLen = strlen(token_strings[i]);
				if(strLen>0)
				{
					bool bMatch = strncmp(token_strings[i], token.cptr, strLen) == 0;

					// is this a better match then
					// what we've found so far?
					if(bMatch && (matchedIndex==-1 || strLen>matchedLen))
					{
						matchedIndex = i;
						matchedLen = strLen;
					}
				}
			}

			if(matchedIndex==-1)
			{
				token.str_len = 1;
				token.token_type = UNKNOWN;
				res = true;
			}
			else
			{
				token.str_len = matchedLen;
				token.token_type = (ASC_TOKEN) matchedIndex;
				res = true;
			}
		}
	}

	return res;
}

int CScriptParser::token_to_integer()
{
	assert(token.token_type == INTEGER);

	int n=0; int res=0;
	if(n<token.str_len && token.cptr[n]=='-')
		++n;
	while(n<token.str_len && isdigit(token.cptr[n]))
	{

		res *= 10;
		res += (int) char_to_digit(token.cptr[n]);
		++n;
	}

	assert(n==token.str_len);
	return (token.cptr[0]=='-') ? -res : res;
}

float CScriptParser::token_to_real()
{
	assert(token.token_type == REAL_VALUE || token.token_type == INTEGER);

	int n=0; float res=0;
	if(n<token.str_len && token.cptr[n]=='-')
		++n;
	while(n<token.str_len && isdigit(token.cptr[n]))
	{

		res *= 10;
		res += (float) char_to_digit(token.cptr[n]);
		++n;
	}

	if(n!=token.str_len || token.token_type!=INTEGER)
	{
		++n;	// skip the dot
		assert(n<token.str_len && token.cptr[n-1]=='.' && token.token_type==REAL_VALUE);
		
		float div = 10.0f;
		while(n<token.str_len && isdigit(token.cptr[n]))
		{
			res += ((float) char_to_digit(token.cptr[n])/div);
			div *= 10.0f;
			++n;
		}
	}

	assert(n==token.str_len);
	return (token.cptr[0]=='-') ? -res : res;
}

int CScriptParser::char_to_digit(const char ch)
{
	// fast
	//return ch-'0';

	// safe
	int val;
	switch(ch)
	{
		case '0':
			val = 0;
			break;
		case '1':
			val = 1;
			break;
		case '2':
			val = 2;
			break;
		case '3':
			val = 3;
			break;
		case '4':
			val = 4;
			break;
		case '5':
			val = 5;
			break;
		case '6':
			val = 6;
			break;
		case '7':
			val = 7;
			break;
		case '8':
			val = 8;
			break;
		case '9':
			val = 9;
			break;
		default:
			assert(false);
	}

	// might do a safefast version using an array

	return val;
}

void CScriptParser::RemoveWhiteSpace()
{
	// remove
	do
	{
		if( (token.cptr+1<(pScript+szeScript)) && token.cptr[0]=='/' && token.cptr[1]=='/' )
			while( (token.cptr < (pScript+szeScript)) && token.cptr[0]!='\n' )
				++token.cptr;
		
		// remove classic whitespace
		while(
			  (token.cptr < (pScript+szeScript)) &&
			  (token.cptr[0]=='\r' || token.cptr[0]=='\t' || token.cptr[0]=='\n' || token.cptr[0]==' ')
			  )
		{
			if(token.cptr[0]=='\n')
				++m_LineNumber;
			++token.cptr;
		}
		
		
	}	// We done? (same as initial test)
	while( (token.cptr+1<(pScript+szeScript)) && token.cptr[0]=='/' && token.cptr[1]=='/' );
}

void CScriptParser::PrintError(const char err_msg[])
{
	const int32 len_str = 12;
	char error_in_script[len_str+1];
	strncpy(error_in_script, token.cptr, len_str);
	error_in_script[len_str] = '\0';
	printf("%s, error at line %d, %s...\n", err_msg, m_LineNumber, error_in_script);
}

