#include <string.h>
#include <math.h>
#include <float.h>
#include "assert.h"
#include "tcbctrl.h"

#define PI			3.1415926f
#define TWOPI		(PI*2.f)
#define HALFPI		(PI*0.5f)
#define COS_EPSILON 1e-6f
#define EPSILON		1e-4f

// initial declerations
Quat QuatLogN(Quat q_in);
Quat QuatExp(Quat q);
Quat QuatAbs(Quat q);
Quat QuatLnDif(const Quat& p, const Quat& q);
Quat SlerpBigQuat(const Quat& p, const Quat& q, float t);
Quat QuadrupleSlerpQuat(const Quat& p, const Quat& a, const Quat &b, const Quat& q, float t);
Quat AAToQuad(Quat angleaxis);




/************************************************************
 *
 *	CTCBAnimCtrl  class
 *
 *	TCB Animation controller
 *
 ************************************************************/

CTCBAnimCtrl::CTCBAnimCtrl()
{
	m_iNumKeys		= 0;

	m_fFrameStart	= 0.f;
	m_fFrameEnd		= 0.f;

	m_pKeys			= 0;
}


CTCBAnimCtrl::~CTCBAnimCtrl()
{
	m_iNumKeys		= 0;

	m_fFrameStart	= 0.f;
	m_fFrameEnd		= 0.f;

	m_pKeys			= 0;
}


// API functions
void CTCBAnimCtrl::EvalPos(float time, Vec3 *vPos)
{
	if(m_iNumKeys == 1)
		*vPos = m_pKeys[0].Vec;
	else
	{
		// evaluate keyStart, keyEnd, blendFactor
		int		keyStart;
		int		keyEnd;
		float	blendFactor;
		
		GetInterpKey(time, keyStart, keyEnd, blendFactor);
		
		if(keyStart != keyEnd) 
			blendFactor = Ease(blendFactor, m_pKeys[keyStart].easefrom, m_pKeys[keyEnd].easeto);

		Key key0 = m_pKeys[keyStart];
		Key key1 = m_pKeys[keyEnd];
		
		// Compute Hermite bases
		
		float bf0,bf1,a;
		
		bf0 = blendFactor*blendFactor;
		bf1 = bf0*blendFactor;
		a  = 2.0f * bf1 - 3.0f*bf0;
		
		*vPos = key0.Vec * (1.0f + a) + 
				key1.Vec * (-a) + 
				key0.tanout * (blendFactor - 2.0f * bf0 + bf1) + 
				key1.tanin * (-bf0 + bf1);
	}
}


void CTCBAnimCtrl::EvalRot(float time, Quat *qRot)
{
	if(GetNumKeys() == 1)
		*qRot = AAToQuad(m_pKeys[0].angleaxis);
	else
	{
		// evaluate keyStart, keyEnd, blendFactor
		float	s,v;
		int		keyStart;
		int		keyEnd;
		float	blendFactor;
		
		GetInterpKey(time, keyStart, keyEnd, blendFactor);

		if(keyStart!=keyEnd) 
			blendFactor = Ease(blendFactor, m_pKeys[keyStart].easefrom, m_pKeys[keyEnd].easeto);

		Key key0 = m_pKeys[keyStart];
		Key key1 = m_pKeys[keyEnd];
		
		float omega = key1.angleaxis.s*.5f;
		float numRevs = 0.0f;					// stands for how many revolutions
		Quat r,pp,qq;

		if (omega<PI-EPSILON) 
		{ 
			*qRot = QuadrupleSlerpQuat(key0.quatfromaxis,key0.A,key1.B,key1.quatfromaxis,blendFactor); 
			return; 
		}

		while (omega > (PI-EPSILON)) 
		{  
			omega -= PI; 
			numRevs += 1.0f; 
		}
		
		if (omega<0.0f) 
			omega = 0.0f;

		s = blendFactor*key1.angleaxis.s/PI;
		
		if (s < 1.0f)
		{
			pp = key0.quatfromaxis*Quat(key1.angleaxis.V,0.0f);
			r = QuadrupleSlerpQuat(key0.quatfromaxis,key0.A,pp,pp,s); 	// in first 90 degrees
		}
		else 
		if ( ( v = s + 1.0f - 2.0f*(numRevs+(omega/PI)) ) <= 0.0f)  
		{
			// middle part, on great circle(p,q)
			while (s >= 2.0f) 
				s -= 2.0f;
			
			pp = key0.quatfromaxis*Quat(key1.angleaxis.V,0.0f);
			r = Slerp(key0.quatfromaxis,pp,s);
		}
		else 
		{ 	  
			// in last 90 degrees
			qq = -key1.quatfromaxis*Quat(key1.angleaxis.V,0.0f);
			r= QuadrupleSlerpQuat(qq,qq,key1.B,key1.quatfromaxis,v);
		}

		*qRot = r;
	}
}


void CTCBAnimCtrl::UpdateDerivates()
{
	int i;

	// sort keys by frame time
	for(int k=0; k<m_iNumKeys-1; k++)
		for(int m=k+1; m<m_iNumKeys; m++)
			if(m_pKeys[k].frame > m_pKeys[m].frame)
			{
				Key tmpkey = m_pKeys[m];
				m_pKeys[m] = m_pKeys[k];
				m_pKeys[k] = tmpkey;
			}
	
	if(m_iNumKeys<2)
		return;

	if (m_iNumKeys==2) 
	{
		Vec3	dp;
		float	t0, t1;
		
		t0 = 1.0f - m_pKeys[0].tens;
		t1 = 1.0f - m_pKeys[1].tens;

		dp = m_pKeys[1].Vec - m_pKeys[0].Vec;
		
		m_pKeys[0].tanout = t0 * dp;
		m_pKeys[1].tanin  = t1 * dp; 
		return;
	}

	float tm,cm,cp,bm,bp,tmcm,tmcp,ksm,ksp,kdm,kdp,c;
	float dt,fp,fn;

	for (i = 1;i<m_iNumKeys-1;i++)
	{
	
		/* fp,fn apply speed correction when continuity is 0.0 */
		if (i==0) 
		{
			float dtb = (m_fFrameEnd-m_pKeys[m_iNumKeys-1].frame + m_pKeys[0].frame-m_fFrameStart);
			dt = .5f * (dtb + m_pKeys[1].frame - m_pKeys[0].frame);
			fp = ((dtb)) / dt;
			fn = ((m_pKeys[1].frame - m_pKeys[0].frame)) / dt;
		}
		else if (i==m_iNumKeys-1) 
		{
			float dta = (m_fFrameEnd-m_pKeys[m_iNumKeys-1].frame + m_pKeys[0].frame-m_fFrameStart);
			dt = .5f * (dta + m_pKeys[m_iNumKeys-1].frame-m_pKeys[m_iNumKeys-2].frame);
			fp = ((dta)) / dt;
			fn = ((m_pKeys[m_iNumKeys-1].frame - m_pKeys[m_iNumKeys-2].frame)) / dt;
		} 
		else 
		{
			dt = .5f * (m_pKeys[i+1].frame- m_pKeys[i-1].frame);
			fp = ((m_pKeys[i].frame - m_pKeys[i-1].frame)) / dt;
			fn = ((m_pKeys[i+1].frame - m_pKeys[i].frame)) / dt;
		}

		c  = fabsf( m_pKeys[i].cont );
		fp = fp + c - c * fp;
		fn = fn + c - c * fn;
		cm = 1.0f - m_pKeys[i].cont; 	
		tm = 0.5f * ( 1.0f - m_pKeys[i].tens) ;
		cp = 2.0f - cm;
		bm = 1.0f - m_pKeys[i].bias;
		bp = 2.0f - bm;      
		tmcm = tm*cm;	tmcp = tm*cp;
		ksm = tmcm*bp*fp;	ksp = tmcp*bm*fp;
		kdm = tmcp*bp*fn; 	kdp = tmcm*bm*fn;
		
		Vec3 * prev_vec = 0;
		Vec3 * cur_vec = 0;
		Vec3 * next_vec = 0;

		if (i==0)
		{
			prev_vec = &m_pKeys[m_iNumKeys-1].Vec;
			cur_vec = &m_pKeys[i].Vec;
			next_vec = &m_pKeys[i+1].Vec;
		}
		else if (i==m_iNumKeys-1)
		{
			prev_vec = &m_pKeys[i-1].Vec;
			cur_vec = &m_pKeys[i].Vec;
			next_vec = &m_pKeys[0].Vec;
		}
		else
		{
			prev_vec = &m_pKeys[i-1].Vec;
			cur_vec = &m_pKeys[i].Vec;
			next_vec = &m_pKeys[i+1].Vec;
		}

		Vec3 delm, delp;

		delm = *cur_vec - *prev_vec;
		delp = *next_vec - *cur_vec;
		m_pKeys[i].tanin  = ksm*delm + ksp*delp;
		m_pKeys[i].tanout = kdm*delm + kdp*delp;
	}

	// Compute the First key's derivates
	m_pKeys[0].tanout = 0.5f * (1.0f - m_pKeys[0].tens) * (3.0f * (m_pKeys[1].Vec - m_pKeys[0].Vec) - m_pKeys[1].tanin);

	// Compute the Last key's derivates
	m_pKeys[m_iNumKeys-1].tanin = -0.5f * (1.0f - m_pKeys[m_iNumKeys-1].tens) * 
										(3.0f * (m_pKeys[m_iNumKeys-2].Vec - m_pKeys[m_iNumKeys-1].Vec) + m_pKeys[m_iNumKeys-2].tanout);
}

void CTCBAnimCtrl::UpdateQuatDerivates()
{
	int n = m_iNumKeys;
	int i;

	// sort keys by frame time
	for(int k=0; k<n-1; k++)
		for(int m=k+1; m<n; m++)
			if(m_pKeys[k].frame > m_pKeys[m].frame)
			{
				Key tmpkey = m_pKeys[m];
				m_pKeys[m] = m_pKeys[k];
				m_pKeys[k] = tmpkey;
			}

	for (i = 0; i < n; ++i)
	{
		Key & key = m_pKeys[i];

		// Derive the absolute quat from angle/axis	
		key.angleaxis = QuatAbs(key.angleaxis);
		
		if(key.angleaxis.V != Vec3(0,0,0))
			key.angleaxis.V = Normalize(key.angleaxis.V);

		key.quatfromaxis = AAToQuad(key.angleaxis);

		if (i!=0) 
			key.quatfromaxis = m_pKeys[i-1].quatfromaxis * key.quatfromaxis;
	}

	
	// IM HERE !!!
	// Then calc the derivs	
	for (i=0; i<n; i++)
	{
		Quat qprev, qnext, q;
		Quat  qp, qm, qa, qb, qae, qbe;	
		float tm,cm,cp,bm,bp,tmcm,tmcp,ksm,ksp,kdm,kdp,c;
		float dt,fp,fn;
			
		if (i>0)
		{
			if (m_pKeys[i].angleaxis.s > TWOPI-.00001f)
			{
				q.V = m_pKeys[i].angleaxis.V;
				q.s = 0.0f;
				qm = QuatLogN(q);
			}
			else
			{
				if (i==0) qprev = m_pKeys[n-1].quatfromaxis;
				else qprev = m_pKeys[i-1].quatfromaxis;
				if (Dot(qprev, m_pKeys[i].quatfromaxis) < 0.0f) qprev = -qprev;
				qm = QuatLnDif(qprev,m_pKeys[i].quatfromaxis);
			}
		}

		if (i<n-1) 
		{
			int i1 = i==n-1 ? 0 : i+1;
			if (m_pKeys[i1].angleaxis.s > TWOPI-.00001f) 
			{
				q.V = m_pKeys[i1].angleaxis.V;
				q.s = 0.0f;
				qp = QuatLogN(q);			
			}
			else 
			{ 
				qnext = m_pKeys[i1].quatfromaxis;			
				if (Dot(qnext, m_pKeys[i].quatfromaxis) < 0.0f) qnext = -qnext;
				qp = QuatLnDif(m_pKeys[i].quatfromaxis, qnext);			
			}
		}
		
		if (i==0 )   qm = qp;
		if (i==n-1) qp = qm;

		fp = fn = 1.0f;
		cm = 1.0f - m_pKeys[i].cont;
		if (i>0 && i<n-1) 
		{
			if (i==0) 
			{
				float dtb = (m_fFrameEnd-m_pKeys[n-1].frame + m_pKeys[0].frame-m_fFrameStart);
				dt = .5f * (dtb + m_pKeys[1].frame-m_pKeys[0].frame);
				fp = ((dtb)) / dt;
				fn = ((m_pKeys[1].frame - m_pKeys[0].frame)) / dt;
			}
			else 
			if (i==n-1) 
			{
				float dta = (m_fFrameEnd-m_pKeys[n-1].frame + m_pKeys[0].frame-m_fFrameStart);
				dt = .5f * (dta + m_pKeys[n-1].frame-m_pKeys[n-2].frame);
				fp = ((dta)) / dt;
				fn = ((m_pKeys[n-1].frame - m_pKeys[n-2].frame)) / dt;
			} 
			else 
			{
				dt = 0.5f * (m_pKeys[i+1].frame - m_pKeys[i-1].frame);
				fp = (m_pKeys[i].frame - m_pKeys[i-1].frame)/dt;
				fn = (m_pKeys[i+1].frame - m_pKeys[i].frame)/dt;
			}
			c = fabsf(m_pKeys[i].cont);
			fp = fp + c - c * fp;
			fn = fn + c - c * fn;
		}

		tm = .5f*(1.0f - m_pKeys[i].tens);
		cp = 2.0f - cm;
		bm = 1.0f - m_pKeys[i].bias;
		bp = 2.0f - bm;      
		tmcm = tm * cm;	
		tmcp = tm * cp;
		ksm  = 1.0f - tmcm * bp * fp;	
		ksp  = -tmcp * bm * fp;
		kdm  = tmcp * bp * fn; 	
		kdp  = tmcm * bm * fn - 1.0f;
		
		qa.V = .5f * (kdm * qm.V + kdp * qp.V);
		qb.V = .5f * (ksm * qm.V + ksp * qp.V);
		qa.s = .5f * (kdm * qm.s + kdp * qp.s);
		qb.s = .5f * (ksm * qm.s + ksp * qp.s);

		qae = QuatExp(qa);
		qbe = QuatExp(qb);
		
		m_pKeys[i].A = m_pKeys[i].quatfromaxis * qae;
		m_pKeys[i].B = m_pKeys[i].quatfromaxis * qbe;	
	}
}



// small utility function
void CTCBAnimCtrl::GetInterpKey(float time, int &keyStart, int &keyEnd, float& blendFactor)
{

	assert( m_iNumKeys>1 );

	if (time <= m_pKeys[0].frame) 
	{
		keyStart	= 0;
		keyEnd		= 1;
		blendFactor = 0.0f;
	} 
	else 
	if (time >= m_pKeys[m_iNumKeys-1].frame) 
	{
		keyStart	= m_iNumKeys-2;
		keyEnd		= m_iNumKeys-1;
		blendFactor	= 1.0f;
	} 
	else
	{
		int i = 0;

		while(i<m_iNumKeys-1 && time>m_pKeys[i+1].frame)
			++i;

		assert(i>=0);

		keyStart = i;
		keyEnd	 = i+1;

		float du = m_pKeys[keyEnd].frame-m_pKeys[keyStart].frame;
		if (du > 0.0f) 
		{
			blendFactor = (time-m_pKeys[keyStart].frame)/du;
		} 
		else 
		{
			blendFactor = 0.0f;
		}			
	}
}

float CTCBAnimCtrl::Ease(float u, float a, float b) 
{
	float k;
	float s = a + b;
	
	if (u==0.0f || u==1.0f) return u;
	
	if (s == 0.0f) 
		return u;

	if (s > 1.0f) 
	{
		a = a/s;
		b = b/s;
	}

	k = 1.f / (2.0f - a - b);
	
	if (u < a) 
		return ((k/a)*u*u);
	else if (u < 1.0f - b) 
		return (k*(2.0f*u - a));
	else
	{
		u = 1.0f - u;
		return (1.0f - (k/b) *u*u);
	}
}

////////////////////////////////////////////////////////////////////////////
///////////////// extended quaternion math functions ///////////////////////
////////////////////////////////////////////////////////////////////////////


//  Natural logarithm of UNIT quaternion 
Quat QuatLogN(Quat q_in)
{
	Quat qq;

	float scale = Length(q_in.V);
	float theta = atan2f(scale, q_in.s);
	
	if (scale > 0.0)
		scale = theta / scale;
	
	qq.V = scale * q_in.V;
	qq.s = 0.0f;
	
	return qq;
}

Quat QuatExp(Quat q)
{
	Quat qq;

	float scale = 1.0;
	float theta = Length(q.V);
	
	if (theta>EPSILON) 
		scale = sinf(theta) / theta;
	
	qq.V = scale * q.V;
	qq.s = cosf(theta);
	return qq;
}

Quat QuatAbs(Quat q)
{
	Quat r = q;
	if (q.s < 0.0f)
	{
		r.s = -r.s;
		r.V  = -r.V;
	}

	return r;
}

Quat divide(const Quat &Qa, const Quat &Qb)
{
	Quat qq;

    float xx = Qb.V.x;
    float yy = Qb.V.y;
    float zz = Qb.V.z;
    float ww = Qb.s;
	float l =	(xx*xx + yy*yy + zz*zz + ww*ww);
	
	if (l == 0.f) 
		l = 1.f;
	
	float norminv = 1.f / l;

	qq.V.x = (-xx * norminv);
	qq.V.y = (-yy * norminv);
	qq.V.z = (-zz * norminv);
	qq.s = (ww * norminv);
	return qq * Qa;
}

Quat QuatLnDif(const Quat& p, const Quat& q)
{
	return(QuatLogN(divide(q, p)));
}

Quat AAToQuad(Quat angleaxis)
{
	float omega, s;
	omega = angleaxis.s*0.5f;
	s = (float)sin(omega);
	return	(Quat(Vec3(s*angleaxis.V.x, s*angleaxis.V.y, s*angleaxis.V.z),(float)cos(omega)));
}

Quat SlerpBigQuat(const Quat& p, const Quat& q, float t)
{
	float omega,cosom,sinom,sclp,sclq;
	
	Quat qt;
	cosom = Dot(p, q);//p.V.x * q.V.x + p.V.y*q.V.y + p.V.z*q.V.z + p.s*q.s;
	
	// the default quaternion slerp module
	if ( (1.0f + cosom) > COS_EPSILON)
	{
		if ( (1.0f - cosom) > COS_EPSILON ) 
		{ 
			omega = acosf(cosom);
			sinom = sinf(omega);

			sclp = sinf((1.0f - t)*omega) / sinom;
			sclq = sinf(t*omega) / sinom;
		}
		else 
		{	// lerp is fine in this case
			sclp = 1.0f - t;
			sclq = t;
		}

		qt.V = sclp*p.V + sclq*q.V;
		qt.s = sclp*p.s + sclq*q.s;
	}
	else
	{
		// extended quaternion slerp module
		t *= 2.0f;
		qt.V.x = -p.V.y;  qt.V.y = p.V.x;  // qt is orthogonal to p 
		qt.V.z = -p.s;  qt.s = p.V.z;
		sclp = sinf((1.0f - t) * HALFPI);
		sclq = sinf(t * HALFPI);
		qt.V = sclp*p.V + sclq*qt.V;
		qt.s = sclp*p.s + sclq*qt.s;
	}

	return qt;
}

Quat QuadrupleSlerpQuat(const Quat& p, const Quat& a, const Quat &b, const Quat& q, float t)
{ 
	float slerpFactor = 2.0f * (1.0f - t) * t;
    return(SlerpBigQuat(SlerpBigQuat(p,q,t),SlerpBigQuat(a,b,t),slerpFactor));
}